/*
 * Copyright 2008-2010 Fabien Coeurjoly <fabien.coeurjoly@wanadoo.fr>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  The names of contributors may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 */

#include <proto/asl.h>
#include <clib/macros.h>

#include "gui/interface.h"
#include "gui.h"

/******************************************************************
 * urlgroupclass
 *****************************************************************/

struct URLGroupData
{
	APTR BT_OK;
	APTR BT_Cancel;
	APTR ST_URL;
};

DEFNEW(URLGroup)
{
	APTR BT_OK, BT_Cancel, ST_URL;

	obj = DoSuperNew(cl, obj,
		Child,
			HGroup,
				Child, MUI_NewObject(MUIC_Dtpic, MUIA_Dtpic_Name, "PROGDIR:images/url.png", End,
				Child, ST_URL = String("", 8192),
				End,
		Child,
			HGroup,
				Child, BT_OK = KeyButton("Ok", 'o'),
				Child, HSpace(0),
				Child, BT_Cancel = KeyButton("Cancel", 'c'),
				End,

		TAG_DONE
	);

	if (obj)
	{
		struct URLGroupData *data = INST_DATA(cl, obj);

		data->BT_OK = BT_OK;
		data->BT_Cancel = BT_Cancel;
		data->ST_URL = ST_URL;

		set(data->ST_URL, MUIA_CycleChain, 1);

		DoMethod(data->BT_OK, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_URLGroup_Open);

		DoMethod(data->BT_Cancel, MUIM_Notify, MUIA_Pressed, FALSE,
				 obj, 1, MM_URLGroup_Cancel);

		DoMethod(data->ST_URL, MUIM_Notify, MUIA_String_Acknowledge, MUIV_EveryTime,
				 obj, 1, MM_URLGroup_Open);
	}

	return (ULONG)obj;
}

DEFTMETHOD(URLGroup_Open)
{
	struct URLGroupData *data = INST_DATA(cl, obj);

	STRPTR url = (STRPTR) getv(data->ST_URL, MUIA_String_Contents);

	DoMethod(mygui->maingroup, MM_MPlayerGroup_Open, MV_MPlayerGroup_Open_Stream, url);

	set(mygui->urlwindow, MUIA_Window_Open, FALSE);

	return 0;
}

DEFTMETHOD(URLGroup_Cancel)
{
	set(mygui->urlwindow, MUIA_Window_Open, FALSE);

	return 0;
}

DEFMMETHOD(Show)
{
	struct URLGroupData *data = INST_DATA(cl, obj);

	ULONG rc = DOSUPER;

	set(mygui->urlwindow, MUIA_Window_ActiveObject, data->ST_URL);

	return rc;
}

BEGINMTABLE2(urlgroupclass)
DECNEW(URLGroup)
DECMMETHOD(Show)
DECTMETHOD(URLGroup_Open)
DECTMETHOD(URLGroup_Cancel)
ENDMTABLE

DECSUBCLASS_NC(MUIC_Group, urlgroupclass, URLGroupData)
