/* 
 *  ao_ahi.c
 *  libao2 module for MPlayer MorphOS 
 *  using AHI/low level API
 *  Writen by DET Nicolas
 *  (C) Genesi 2003
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <osdep/timer.h>
                  
#include "mp_msg.h"

// NOTE: all OS Specific stuff have not been protected by #ifndef ... #else .. #endif
// because il you compile ahi, you are using on MorphOS (or AmigaOS)

// Some OS Specific include
#ifndef __AMIGAOS4A__
#include <proto/alib.h>
#include <hardware/byteswap.h>
#else
#include "amigaword.h"
#include <clib/alib_protos.h>
#endif
#include <proto/ahi.h>
#include <proto/exec.h>
#include <proto/dos.h>

#include <dos/dostags.h>
#include <dos/dos.h>

#ifdef __amigaos4__
struct AHIIFace  *IAHI   = NULL;
#endif



// End OS Specific

#include "audio_out.h"
#include "audio_out_internal.h"

#include "../libaf/af_format.h"


// DEBUG
#define KPRINTF KPrintF

#if 0
#include <clib/debug_protos.h>
#define DEBUG(x) x
#else
#define DEBUG(x) 
#endif

/****************************************************************************/

//set and define by mplayer.c

struct Library *  AHIBase = NULL;

static struct MsgPort *     AHIMsgPort  = NULL;     // The msg port we will use for the ahi.device
static struct AHIRequest *  AHIio       = NULL;     // First AHIRequest structure
static BYTE AHIDevice                   = -1;       // Is our AHI device opened ? 1 -> no, 0 -> Yes
static ULONG AHIType;

#define PLAYERTASK_NAME       "MPlayer audio task"
#define PLAYERTASK_PRIORITY   2


static struct Task   	*PlayerTask_Task = NULL;
static struct Task      *MainTask = NULL;           // Pointer to the main task;

// some AmigaOS Signals
static BYTE sig_taskend		=-1;
static BYTE sigsound			=-1;

static struct AHIAudioCtrl *actrl = NULL;

static LONG AHI_Volume=0x10000;
static struct AHISampleInfo Sample0= { 0, NULL, NULL };
static struct AHISampleInfo Sample1= { 0, NULL, NULL };

static BOOL 	AHI_SampleLoaded = FALSE;

// Buffer management
static struct SignalSemaphore LockSemaphore;
static ULONG buffer_get=0; // place where we have to get the buffer to play
static ULONG buffer_put=0; // Place where we have to put the buffer we will play
static int ChunkLen=0; // Size of each chuck we will play
static int TotalLen=0; // Total size of the buffer

static unsigned int written_byte=0;
static BYTE *buffer=NULL;    // Our audio buffer

static  BOOL DBflag=FALSE;  // double buffer flag
static  BOOL running = TRUE; // switching to FALSE will stop and kill the PlayerTask

typedef enum
{
  CO_NONE,
  CO_8_U2S,
  CO_16_U2S,
  CO_16_LE2BE,
  CO_16_LE2BE_U2S
} convop_t;
convop_t convop;

//Custom function internal to this driver
inline static void cleanup(void); // Clean !

static ao_info_t info = 
{
#ifdef __MORPHOS__
    "AHI audio output using low-level API (AmigaOS)",
#else
   "AHI audio output using low-level API (AmigaOS)",
#endif
    "ahi",
    "DET Nicolas",
#ifdef __MORPHOS__
      "AmigaOS4 Rulez :-)"
#else
      "AmigaOS4 Rulez!",
#endif
};

// define the standart functions that mplayer will use.
// see define in audio_aout_internal.h
LIBAO_EXTERN(ahi)

/*******************************************************************/
static void cleanup(void) {
	DEBUG(KPRINTF("Cleanup !\n"));

	if (AHIBase)
	{
		if ( AHI_SampleLoaded ) {
			DEBUG(KPRINTF("Unloading Sample\n"));
			AHI_UnloadSound ( 0, actrl);
        	AHI_UnloadSound ( 1, actrl);
			AHI_SampleLoaded = FALSE;
		}

		if (Sample0.ahisi_Address) {
        	DEBUG(KPRINTF("Freeing Sample buffer\n"));
			FreeVec(Sample0.ahisi_Address);
			Sample0.ahisi_Address = NULL;
		}

		if (Sample1.ahisi_Address) {
			FreeVec(Sample1.ahisi_Address);
			Sample1.ahisi_Address = NULL;
		}

		if (actrl) {
        	DEBUG(KPRINTF("Freeing Audio Hardware\n"));
	    	AHI_FreeAudio (actrl);
			actrl = NULL;
		}
		AHIBase = NULL;
	}

   	if (! AHIDevice ) {
		DEBUG(KPRINTF("Closing device\n"));
		CloseDevice ( (struct IORequest *) AHIio);
		AHIDevice = 1;       // 1 -> Not opened !
	}

	if (AHIio) {
		DEBUG(KPRINTF("Deleting IORequest\n"));
		DeleteIORequest ( (struct IORequest *) AHIio);	
		AHIio = NULL;
	}

	if (AHIMsgPort) {
		DEBUG(KPRINTF("Deleting AHIMsgPort\n"));
		DeleteMsgPort(AHIMsgPort);
		AHIMsgPort = NULL;
		}

	DEBUG(KPRINTF("Releasing interface\n"));
	if (IAHI) DropInterface((struct AHIIFace*) IAHI);	

	if (buffer) {
      DEBUG(KPRINTF("Freeing Buffer\n"));
		FreeVec(buffer );
		buffer = NULL;
	}

	if (sigsound != -1) {
		DEBUG(KPRINTF("Freeing signal sigsound\n"));
      FreeSignal (sigsound );
      sigsound =-1;
	}
	if (sig_taskend != -1) {
		DEBUG(KPRINTF("Freeing signal taskend\n"));
      FreeSignal ( sig_taskend );   
      sig_taskend=-1; 
	}

	DEBUG(KPRINTF("The end\n"));
  
}


/******************** PLAYER TASK ********************/
static void PlayerTask (void) {

DEBUG(KPRINTF("TASK: Starting main loop\n"));
while(running) {

	Wait ( 1L << sigsound);
	if (!running) break;

	ObtainSemaphore(&LockSemaphore);
	DEBUG(KPRINTF("TASK: Semaphore obtained\n"));

	if ( ! written_byte ) {
		if ( DBflag ) memset( Sample1.ahisi_Address, 0, ChunkLen);
		else memset( Sample0.ahisi_Address, 0, ChunkLen);
        DEBUG(KPRINTF("TASK: Semaphore released (Buffer EMPTY)\n"));
        ReleaseSemaphore(&LockSemaphore);
		continue;
	}

	// Copy data in the good sample buffer
	// if ( DBflag ) // Sound 0 has been launched -> filling buffer 1
	// else ...
	if ( DBflag ) memcpy( Sample1.ahisi_Address, buffer + buffer_get, ChunkLen);
	else memcpy( Sample0.ahisi_Address, buffer + buffer_get, ChunkLen);
   
	// Update some stuff
	written_byte   -= ChunkLen;
	buffer_get     += ChunkLen;
	buffer_get     %= TotalLen;

	DEBUG(KPRINTF("TASK: Semaphore released (next line)\n"));
	ReleaseSemaphore(&LockSemaphore);
}

Signal( (struct Task *) MainTask , 1L<< sig_taskend );
DEBUG(KPRINTF("TASK: sig_taskend -> break\n"));
}

/******************** SOUND FUNC ********************/

static void SoundFunc (void) {
   //struct Hook *hook = (struct Hook *)REG_A0;
   //struct AHIAudioCtrl * actrl = (struct AHIAudioCtrl *)REG_A2;
   //struct AHISoundMessage * smsg = (struct AHISoundMessage *)REG_A1;

   if( (DBflag = !DBflag) ) // Flip and test
      AHI_SetSound(0,1,0,0,actrl,NULL);
   else
      AHI_SetSound(0,0,0,0,actrl,NULL);

   DEBUG(KPRINTF("Hook : sigsound sent to playtask\n"));
   Signal( (struct Task *) PlayerTask_Task, 1L << sigsound );
}

#ifndef __AMIGAOS4__
struct EmulLibEntry GATE_SoundFunc = {
   TRAP_LIB, 0, (void (*)(void))SoundFunc
};

struct Hook SoundFunc_Hook = {
   {NULL, NULL},
   (APTR) &GATE_SoundFunc,
   NULL,
   NULL,
};
#else
struct Hook SoundFunc_Hook =
{
   {NULL, NULL},
   (HOOKFUNC)SoundFunc,
   NULL,
   NULL
};
#endif

/***************************** CONTROL *******************************/
// to set/get/query special features/parameters
static int control(int cmd, void *arg) 
{
	switch (cmd) 
	{
		case AOCONTROL_GET_VOLUME:
		{
			ao_control_vol_t* vol = (ao_control_vol_t*)arg;
			vol->left = vol->right =  ( (float) (AHI_Volume >> 8 ) ) / 2.56 ;
			DEBUG(KPRINTF("vol->left=%f\n", vol->left));
			return CONTROL_OK;
		}

		case AOCONTROL_SET_VOLUME:
		{
			float diff;
			ao_control_vol_t* vol = (ao_control_vol_t*)arg;
			diff = (vol->left+vol->right) / 2;
			AHI_Volume =  ( (int) (diff * 2.56) ) << 8 ;
			AHI_SetVol(0, AHI_Volume, 0x8000L, actrl, AHISF_IMM);
			DEBUG(KPRINTF("AHI_VOLUME=%x diff=%f\n", AHI_Volume, diff));
			return CONTROL_OK;
		}
	}
	return CONTROL_UNKNOWN;
}

/***************************** INIT **********************************/
// open & setup audio device
// return: 1=success 0=fail
static int init(int rate,int channels,int format,int flags){
      // Init ao.data to make mplayer happy :-)  
   	ao_data.channels=      channels;
   	ao_data.samplerate=    rate;
   	ao_data.format=        format;
   	ao_data.bps=           channels*rate;



   	if (format != AF_FORMAT_U8 && format != AF_FORMAT_S8) ao_data.bps*=2; // And 24bit sound ???
  
		AHIBase= NULL;

   	DEBUG(KPRINTF("Init AUDIO\n"));

   	convop = CO_NONE;

   	switch ( format ) {

      case AF_FORMAT_U8:
         convop  = CO_8_U2S;
         AHIType = AHIST_S8S;
         break;
      case AF_FORMAT_S8:
         AHIType = AHIST_S8S;
         break;
      case AF_FORMAT_U16_BE:
         convop  = CO_16_U2S;
         AHIType = AHIST_S16S;
         break;
      case AF_FORMAT_S16_BE:
         AHIType = AHIST_S16S;
         break;
      case AF_FORMAT_U16_LE:
         convop  = CO_16_LE2BE_U2S;
         AHIType = AHIST_S16S;
         break;
      case AF_FORMAT_S16_LE:
         convop  = CO_16_LE2BE;
         AHIType = AHIST_S16S;
         break;
      default:
        mp_msg(MSGT_AO, MSGL_WARN, "AHI: Sound format not supported by this driver.\n"); 
		cleanup();
		 return 0; // return fail !
      }


    	DEBUG(KPRINTF("Init AHI\n"));

   	// Init AHI

   	if( ( AHIMsgPort=CreateMsgPort() )) {
      	if ( (AHIio=(struct AHIRequest *)CreateIORequest(AHIMsgPort,sizeof(struct AHIRequest)))) {
         	AHIio->ahir_Version = 4;
         	AHIDevice=OpenDevice(AHINAME, AHI_NO_UNIT,(struct IORequest *)AHIio,NULL);
      	}
   	}
	
   	if(AHIDevice) {
      	DEBUG(KPRINTF("Unable to open %s/0 version 4\n",AHINAME));
      	cleanup();
      	return 0;
   	}

   	AHIBase = (struct Library *) AHIio->ahir_Std.io_Device;
	if(!(IAHI = (struct AHIIFace *)GetInterface(AHIBase, "main", 1L, NULL)))
	{
		DEBUG(KPRINTF("Cannot open AHI Interface!\n");)
		cleanup();
		return 0;
	}
   memset(&LockSemaphore, 0, sizeof(LockSemaphore));
   InitSemaphore(&LockSemaphore);

   // Calculate some size, etc ...
   {
      ULONG playsamples;
      ULONG AudioID = AHI_BestAudioID(
								AHIDB_Stereo,     (AHIType == AHIST_S16S) ? TRUE : FALSE,
								AHIDB_Bits,       (AHIType == AHIST_S16S) ? 16 : 8,
								AHIDB_HiFi,       TRUE,
								AHIDB_Volume,     TRUE,
								AHIDB_Panning,    TRUE,
						        AHIDB_MinMixFreq, rate,
                        TAG_DONE);
      if ( AudioID == AHI_INVALID_ID ) {
         cleanup();
         mp_msg(MSGT_AO, MSGL_ERR, "No good AudioID found.\n");
			return 0;
      }

      mp_msg(MSGT_AO, MSGL_INFO, "AHI: Using 0x%x AudioID.\n", (int) AudioID);

		DEBUG(KPRINTF("AllocAudio\n"));
      if(!( actrl=AHI_AllocAudio(
                   // default mode...
                   //**AHIA_AudioID,    AudioID,
                   AHIA_Channels,   1,
                   AHIA_Sounds,     2,
                   AHIA_SoundFunc,  (ULONG) &SoundFunc_Hook,
                   TAG_DONE) ) ) {
         // Problem
         cleanup();
         mp_msg(MSGT_AO, MSGL_ERR, "Unable to allocate audio hardware.\n");
         return 0;
      }

		DEBUG(KPRINTF("GetAduiAttrs\n"));
      AHI_GetAudioAttrs(AHI_INVALID_ID,
            actrl,
            AHIDB_MaxPlaySamples, (ULONG) &playsamples,
            TAG_DONE);

      // Here, ChunkLen is in samples because ahisi_Length from
      // Samplex want samples
		ChunkLen = playsamples * 3 ;

      Sample0.ahisi_Type= AHIType;
      Sample1.ahisi_Type= AHIType;

      Sample0.ahisi_Length= ChunkLen;
      Sample1.ahisi_Length= ChunkLen;

      // Now just converts ChunkLen is byte
      ChunkLen *= AHI_SampleFrameSize( AHIType );

      //Total buffer size
		TotalLen = ChunkLen * 32;
      mp_msg(MSGT_AO, MSGL_INFO, "AHI: Using %d bytes per chunk and %d Kb for buffer\n", ChunkLen, TotalLen/1024);

      DEBUG(KPRINTF("Tring to Alloc some mem\n"));
      Sample0.ahisi_Address=AllocVec(ChunkLen,MEMF_PUBLIC|MEMF_CLEAR);
      Sample1.ahisi_Address=AllocVec(ChunkLen,MEMF_PUBLIC|MEMF_CLEAR);
   }

   if ( (!Sample0.ahisi_Address) || (!Sample1.ahisi_Address) ) {
      cleanup();
      DEBUG(KPRINTF("Alloc ChunkLen = %d\n", ChunkLen));
      mp_msg(MSGT_AO, MSGL_ERR, "AHI: Not enough memory.\n");
      return 0;
   } 

   if ( ! (buffer = AllocVec ( TotalLen  , MEMF_PUBLIC|MEMF_CLEAR) ) ) {
      cleanup();
      DEBUG(KPRINTF("Alloc AHI_MAXBUFFER failed %d\n", ChunkLen));
      mp_msg(MSGT_AO, MSGL_ERR, "AHI: Not enough memory.\n");
      return 0;
   }

   DEBUG(KPRINTF("Trying to load AHI sound\n"));
   if( (AHI_LoadSound(0, AHIST_DYNAMICSAMPLE, &Sample0, actrl) ) ||
       (AHI_LoadSound(1, AHIST_DYNAMICSAMPLE, &Sample1, actrl) ) ) {
      mp_msg(MSGT_AO, MSGL_ERR, "AHI: Unable to load sound.\n");
      cleanup();
      return 0;
   }

	AHI_SampleLoaded = TRUE;

	MainTask = FindTask(NULL);
	running = TRUE;
	
	DEBUG(KPRINTF("Init Signal\n"));
   	if ( ( sigsound = AllocSignal ( -1 ) )  == -1 ) {
      	cleanup();
      	return 0;
   	}

   	//Allocate our signal
   	if ( ( sig_taskend = AllocSignal ( -1 ) )  == -1 ) {
      	cleanup();
      	return 0;
   	}

	DEBUG(KPRINTF("Creating Task\n"));
#ifndef __AMIGAOS4__
	if ( ! (PlayerTask_Task = NewCreateTask(
					TASKTAG_CODETYPE,    CODETYPE_PPC,
					TASKTAG_PC,          (ULONG) PlayerTask,
					TASKTAG_NAME,        PLAYERTASK_NAME,
					TASKTAG_PRI,         PLAYERTASK_PRIORITY,
					TAG_DONE) ) ) {
#else
	if ( ! (PlayerTask_Task = CreateTask(
				        PLAYERTASK_NAME, PLAYERTASK_PRIORITY, 
					PlayerTask, 
					128*1024, NULL)) ) {
#endif
            DEBUG(KPRINTF("AHI: Unable to create the audio process.\n"));
            mp_msg(MSGT_AO, MSGL_WARN, "AHI: Unable to create the audio process.\n");
            cleanup();
            return 0;
	}
  
	if ( AHI_ControlAudio(actrl,
                        AHIC_Play,TRUE,
                        TAG_DONE) ) {
      mp_msg(MSGT_AO, MSGL_WARN, "AHI: Unable to start playing.\n");
      cleanup();
      return 0;
	}

#if 1
	AHI_SetFreq(0, rate, actrl, AHISF_IMM);
	AHI_SetVol(0, AHI_Volume, 0x8000L, actrl, AHISF_IMM);
	AHI_SetSound(0, 0, 0, 0, actrl, AHISF_IMM);
#else
   AHI_Play(actrl,
                    AHIP_BeginChannel, 0,
                    AHIP_Freq,         rate,
                    AHIP_Vol,          0x10000L,
                    AHIP_Pan,          0x8000L,
                    AHIP_Sound,        0,
                    AHIP_Offset,       0,
                    AHIP_Length,       0,
                    AHIP_EndChannel,   NULL,
                    TAG_DONE);
#endif

   return 1;                                // Everything is ready to go !
}

/***************************** UNINIT ********************************/


// close audio device

static void uninit(int immed)
{
	running = FALSE;
	//ObtainSemaphore(&LockSemaphore);
	//D(kprintf("Uninit\n"));
	if (sig_taskend != -1 && PlayerTask_Task)
	{
		//D(kprintf("Waiting for the PlayerTask \n"));
		Wait(1L << sig_taskend ); // Wait for the PlayerTask to be finished before leaving !
	}

	if (actrl)
	{
		//aci.aeci.ahie_Effect = AHIET_CHANNELINFO|AHIET_CANCEL;
		//AHI_SetEffect(&aci.aeci, actrl);
		AHI_ControlAudio(actrl,
						 AHIC_Play,FALSE,
						 TAG_DONE);
	}

	cleanup();
	//ReleaseSemaphore(&LockSemaphore);
	//D(kprintf("Uninit END\n"));
}



/***************************** RESET **********************************/
// stop playing and empty buffers (for seeking/pause)
static void reset(){
    memset( Sample0.ahisi_Address, 0, ChunkLen);
    memset( Sample1.ahisi_Address, 0, ChunkLen);
    buffer_get=0;
    buffer_put=0;
    written_byte=0;
}

/***************************** PAUSE **********************************/
// stop playing, keep buffers (for pause)
static void audio_pause()
{
AHI_ControlAudio(actrl,
                        AHIC_Play,FALSE,
                        TAG_DONE);
}

/***************************** RESUME **********************************/
// resume playing, after audio_pause()
static void audio_resume()
{
AHI_ControlAudio(actrl,
                        AHIC_Play,TRUE,
                        TAG_DONE);
}

/***************************** GET_SPACE *******************************/
// return: how many bytes can be played without blocking
static int get_space(){
   DEBUG(KPRINTF("AHI_MAXBUFFER - written_byte = %d\n", TotalLen - written_byte));
   return TotalLen - written_byte; // aka total - used
}

/***************************** PLAY ***********************************/
// plays 'len' bytes of 'data'
// it should round it down to outburst*n
// return: number of bytes played
static int play(void* data,int len,int flags){
   // This function just fill a buffer which is played by another task
   int data_to_put;   
   int i, imax;

   DEBUG(KPRINTF("play\n"));
   
	ObtainSemaphore(&LockSemaphore); // Semaphore to protect our variable as we have 2 tasks
   DEBUG(KPRINTF("Semaphore Obtained\n"));

   if ( written_byte == TotalLen ) {
		DEBUG(KPRINTF("Semaphore Released (next line)\n"));
		ReleaseSemaphore(&LockSemaphore);
		return 0;
   }
   
   // Buffer not full -> we put data inside
   data_to_put = buffer_get - buffer_put;
   if ( (data_to_put < 0) || ( (!data_to_put) && ( written_byte != TotalLen ) ) ) data_to_put = TotalLen - buffer_put;
   if ( data_to_put > len) data_to_put = len ;

   DEBUG(KPRINTF("Filling the buffer: data_to_put:%d\n", data_to_put));
   switch (convop)
   {
     case CO_NONE:
       CopyMem( data, buffer + buffer_put, data_to_put);    
       break;

     case CO_8_U2S:
       imax = data_to_put;
       for (i = 0; i < imax ; i++)
       {
         ((BYTE *) (buffer + buffer_put))[i] = ((UBYTE *)data)[i] - 128;
       }
       break;

     case CO_16_U2S:
       imax = data_to_put / 2;
       for (i = 0; i < imax; i++)
       {
         ((WORD *) (buffer + buffer_put))[i] = ((UWORD *)data)[i] - 32768;
       }
       break;

     case CO_16_LE2BE_U2S:
       imax = data_to_put / 2;
       for (i = 0; i < imax; i++)
       {
         ((WORD *) (buffer + buffer_put))[i] = SWAPWORD(((UWORD *)data)[i]) - 32768;
       }
       break;

     case CO_16_LE2BE:
       imax = data_to_put / 2;
       for (i = 0; i < imax; i++)
       {
         ((WORD *) (buffer + buffer_put))[i] = SWAPWORD(((WORD *)data)[i]);
       }
       break;
   }
   
   // update buffer_put
   buffer_put += data_to_put;
   buffer_put %= TotalLen;

   written_byte += data_to_put;
   DEBUG(KPRINTF("Semaphore Released (next line);\n"));
	ReleaseSemaphore(&LockSemaphore); // We have finished to use variables which need protection
  
   return data_to_put;
}

/***************************** GET_DELAY **********************************/
// return: delay in seconds between first and last sample in buffer
static float get_delay(){
	return written_byte / (float) ao_data.bps;
}
